# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import sys
import types
from abc import ABCMeta


class CompoundArgument(object):
    pass


class AttributesBinder(object):

    def __init__(self, obj):
        self.obj = obj

    def __call__(self, attr, val):
        setattr(self.obj, attr, val)


def type_object(obj):
    return hasattr(obj, '__bases__')


def nontype_object(obj):
    return not type_object(obj)


def raise_on_non_function(function):
    if not isinstance(function, types.FunctionType):
        raise MagicTypeError(
            'require FunctionType.',
            function=function,
        )


def raise_on_nontype_object(obj):
    if nontype_object(obj):
        raise MagicTypeError(
            'require type object.',
            obj=obj,
        )


def repr_return(text):
    if sys.version_info.major == 2:  # pragma: no cover
        text = text.encode('utf-8')
    return text


def conditional_repr(obj):
    if isinstance(obj, ABCMeta) and hasattr(obj, 'main_cls'):
        return repr(obj)
    elif type_object(obj):
        return obj.__name__
    else:
        return repr(obj)


from magic_constraints.exception import (
    # MagicSyntaxError,
    MagicTypeError,
)  # noqa



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from magic_constraints.exception import MagicSyntaxError, MagicTypeError


def transform_to_slots(constraints_package, *args, **kwargs):

    class UnFill(object):
        pass

    plen = len(constraints_package.parameters)

    if len(args) > plen:
        raise MagicSyntaxError(
            'argument length unmatched.',
            parameters=constraints_package.parameters,
            args=args,
        )

    slots = [UnFill] * plen
    unfill_count = plen

    # 1. fill args.
    for i, val in enumerate(args):
        slots[i] = val
    unfill_count -= len(args)

    # 2. fill kwargs.
    for key, val in kwargs.items():
        if key not in constraints_package.name_hash:
            raise MagicSyntaxError(
                'invalid keyword argument',
                parameters=constraints_package.parameters,
                key=key,
            )

        i = constraints_package.name_hash[key]
        if slots[i] is not UnFill:
            raise MagicSyntaxError(
                'key reassignment error.',
                parameters=constraints_package.parameters,
                key=key,
            )

        slots[i] = val
        unfill_count -= 1

    # 3. fill defaults if not set.
    # 3.1. deal with the case that default not exists.
    default_begin = constraints_package.start_of_defaults
    if default_begin < 0:
        default_begin = plen
    # 3.2 fill defaults.
    for i in range(default_begin, plen):
        parameter = constraints_package.parameters[i]
        j = constraints_package.name_hash[parameter.name]

        if slots[j] is UnFill:
            slots[j] = parameter.default
            unfill_count -= 1

    # 4. test if slots contains UnFill.
    if unfill_count != 0:
        raise MagicSyntaxError(
            'slots contains unfilled argument(s).',
            parameters=constraints_package.parameters,
            slots=slots,
        )

    return slots


def check_and_bind_arguments(parameters, slots, bind_callback):

    plen = len(parameters)

    for i in range(plen):
        arg = slots[i]
        parameter = parameters[i]

        # check.
        if not parameter.check_argument(arg):
            raise MagicTypeError(
                'argument unmatched.',
                parameter=parameter,
                argument=arg,
            )

        # bind.
        bind_callback(parameter.name, arg)



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from collections import namedtuple
import collections as abc

from funcsigs import signature
from funcsigs import Parameter as SigParameter

from magic_constraints.types import Any
from magic_constraints.exception import (
    MagicSyntaxError,
    MagicTypeError,
)
from magic_constraints.utils import (
    type_object, nontype_object,
    raise_on_nontype_object,
    repr_return, conditional_repr,
)


class Constraint(object):

    def __init__(self, type_, **options):

        raise_on_nontype_object(type_)
        self.type_ = type_

        # key only arguemnt.
        # 1. if nullable, None is accepted.
        self.nullable = options.get('nullable', False)

        # 2. record default value.
        # NOTICE that ReturnType do not support default.
        if 'default' in options:
            self.with_default = True
            self.default = options['default']
        else:
            self.with_default = False
            self.default = None

        # 3. user-defined validator on input argument.
        self.validator = options.get(
            'validator', lambda instance: True,
        )

        # generator serialized string for repr.
        # op 1.
        prefix = self.init_arguments_repr_prefix(type_, **options)
        # common suffix.
        suffix = self._init_arguments_repr_suffix(type_, **options)
        if prefix:
            self._arguments_repr = '{0}, {1}'.format(prefix, suffix)
        else:
            self._arguments_repr = suffix

    def check_argument(self, instance):
        if instance is None:
            if not (self.nullable or issubclass(self.type_, type(None))):
                return False
            else:
                return self.validator(None)
        if not isinstance(instance, self.type_):
            return False
        return self.validator(instance)

    def init_arguments_repr_prefix(self, type_, **options):
        return ''

    def _init_arguments_repr_suffix(self, type_, **options):
        # positional arguments.
        prefix = "type_={type_}".format(
            type_=conditional_repr(type_),
        )
        # keyword-only arguments.
        suffix = ', '.join(map(
            lambda p: "{0}={1}".format(p[0], conditional_repr(p[1])),
            sorted(options.items(), key=lambda p: p[0]),
        ))

        # merge.
        if suffix:
            arguemnt_repr = "{0}, {1}".format(prefix, suffix)
        else:
            arguemnt_repr = prefix

        return arguemnt_repr

    def __repr__(self):
        return repr_return(
            '{cls_name}({arguemnt_repr})'.format(
                cls_name=type(self).__name__,
                arguemnt_repr=self._arguments_repr,
            ),
        )


class Parameter(Constraint):

    def __init__(self, name, type_, **options):
        self.name = name
        super().__init__(type_, **options)

    def init_arguments_repr_prefix(self, type_, **options):
        return 'name={0}'.format(
            conditional_repr(self.name),
        )


class ReturnType(Constraint):

    def __init__(self, type_, **options):
        if 'default' in options:
            raise MagicSyntaxError(
                'default is invalid in ReturnType',
                default=options['default'],
            )
        super().__init__(type_, **options)


def raise_on_non_parameters(parameters):
    if not isinstance(parameters, abc.Sequence):
        raise MagicTypeError(
            'parameter should be Sequence.',
            parameters=parameters,
        )

    for p in parameters:
        if isinstance(p, Parameter):
            continue
        else:
            raise MagicTypeError(
                'require instance of Parameter.',
                p=p,
            )


def raise_on_non_constraints(constraints):
    raise_on_non_parameters(constraints[:-1])
    if not isinstance(constraints[-1], Constraint):
        raise MagicTypeError(
            'require Parameter or ReturnType in the ending of constraints.',
            constraints=constraints,
        )


def check_and_preprocess_parameters(parameters):
    raise_on_non_parameters(parameters)

    name_hash = {}
    start_of_defaults = -1

    idx = 0
    for parameter in parameters:

        if parameter.name in name_hash:
            raise MagicSyntaxError(
                'redefinition of parameter',
                name=parameter.name,
            )

        name_hash[parameter.name] = idx

        if parameter.with_default:
            if start_of_defaults < 0:
                start_of_defaults = idx
        else:
            if start_of_defaults >= 0:
                raise MagicSyntaxError(
                    'Parameter without default should be placed '
                    'in front of all parameters with default.',
                    name=parameter.name,
                )
        idx += 1

    return name_hash, start_of_defaults


def build_constraints_package(constraints):
    if isinstance(constraints[-1], ReturnType):
        parameters = constraints[:-1]
        return_type = constraints[-1]
    else:
        parameters = constraints
        return_type = ReturnType(Any, nullable=True)

    name_hash, start_of_defaults = check_and_preprocess_parameters(parameters)

    ConstraintsPackage = namedtuple(
        'ConstraintsPackage',
        [
            'parameters', 'name_hash', 'start_of_defaults',
            'return_type',
        ],
    )

    return ConstraintsPackage(
        parameters, name_hash, start_of_defaults,
        return_type,
    )


def build_parameter_in_inspection(name, type_, sig_parameter):
    # check the form of parameter.
    if sig_parameter.kind not in [
        SigParameter.POSITIONAL_ONLY,
        SigParameter.POSITIONAL_OR_KEYWORD
    ]:
        raise MagicSyntaxError(
            'supports only POSITIONAL_ONLY '
            'or POSITIONAL_OR_KEYWORD argument.',
            name=name,
            kine=SigParameter.kind,
        )

    annotation = sig_parameter.annotation
    default = sig_parameter.default

    # two cases on annotation.
    if type_ is None:
        if annotation is SigParameter.empty:
            raise MagicSyntaxError(
                'missing annotation',
                name=name,
            )
    else:
        if nontype_object(type_):
            raise MagicTypeError(
                'type_ should be a type object.',
                name=name,
                type_=type_,
            )
        annotation = type_

    if default is SigParameter.empty:
        # without default.
        return Parameter(name, annotation)
    elif default is None:
        # with None as default.
        return Parameter(name, annotation, nullable=True, default=None)
    else:
        # with non-None as default.
        return Parameter(name, annotation, default=default)


def build_return_type_in_inspection(return_type):
    if return_type is None:
        return ReturnType(type(None)),

    elif return_type is SigParameter.empty:
        return ReturnType(Any, nullable=True)

    elif type_object(return_type):
        return ReturnType(return_type)

    else:
        raise MagicTypeError(
            'return_type should be None or type object.',
            return_type=return_type,
        )


def build_constraints_with_annotation(function, skip_first_argument):
    function_sig = signature(function)

    constraints = []
    # 1. parameters.
    for name, sig_parameter in function_sig.parameters.items():
        if skip_first_argument:
            skip_first_argument = False
            continue

        constraints.append(
            build_parameter_in_inspection(name, None, sig_parameter),
        )

    # 2. return type.
    constraints.append(
        build_return_type_in_inspection(function_sig.return_annotation),
    )

    return constraints


def build_constraints_with_given_type_args(
        function, skip_first_argument,
        type_args, return_type=SigParameter.empty):

    argument_sigs = signature(function)

    parameter_length = len(argument_sigs.parameters)
    if skip_first_argument:
        parameter_length -= 1

    if parameter_length != len(type_args):
        raise MagicSyntaxError(
            'length of arguments not match.',
            type_args=type_args,
            function_signature=argument_sigs.parameters,
        )

    constraints = []
    # 1. parameters.
    ti = 0
    for name, sig_parameter in argument_sigs.parameters.items():
        if skip_first_argument:
            skip_first_argument = False
            continue

        constraints.append(
            build_parameter_in_inspection(name, type_args[ti], sig_parameter),
        )
        ti += 1

    # 2. return type.
    constraints.append(
        build_return_type_in_inspection(return_type),
    )

    return constraints



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import pytest

from magic_constraints import *  # noqa


def test_on_init():

    @class_initialization_constraints
    class Case(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
            Parameter('b', list),
            Parameter('c', list, default=[1, 2, 3]),
        ]

        def __init__(self):
            assert self.a == 42
            assert self.b
            assert self.c == [1, 2, 3]

    Case(42, [1])
    with pytest.raises(AssertionError):
        Case(42, [])
    with pytest.raises(AssertionError):
        Case(42, [1], [1, 2, 3, 4])
    Case(42, [1], [1, 2, 3])

    with pytest.raises(SyntaxError):
        Case(1.0)
    with pytest.raises(TypeError):
        Case(42, 1)
    with pytest.raises(TypeError):
        Case(42, [1, 2, 3], 42)


def test_none():

    @class_initialization_constraints
    class Case(object):

        INIT_PARAMETERS = [
            Parameter('a', int, nullable=True, default=None),
            Parameter('b', list, nullable=True, default=None),
        ]

    Case()
    Case(1)
    Case(1, [])
    Case(None)
    Case(None, None)


def test_on_typeobj():

    @class_initialization_constraints
    class Case1(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

    Case1(1)
    with pytest.raises(TypeError):
        Case1(1.0)

    @class_initialization_constraints
    class Case2(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

        def __init__(self):
            assert self.a != 42

    Case2(1)
    Case2(a=1)
    with pytest.raises(AssertionError):
        Case2(42)


def test_class_parameter():

    @class_initialization_constraints
    class Case1(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

    Case1(1)
    Case1(a=1)
    with pytest.raises(SyntaxError):
        Case1(b=1)

    @class_initialization_constraints
    class Case2(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

        def __init__(self):
            assert self.a != 42

    Case2(1)
    Case2(a=1)
    with pytest.raises(AssertionError):
        Case2(42)


def test_nested_type():

    @class_initialization_constraints
    class Case(object):

        INIT_PARAMETERS = [
            Parameter('a', Union[list, tuple]),
        ]

    Case(
        [1, 2],
    )
    Case(
        (1, 2),
    )
    with pytest.raises(TypeError):
        Case(1)

    with pytest.raises(TypeError):

        @class_initialization_constraints
        class Case1(object):

            INIT_PARAMETERS = [
                Parameter('a', Union[list, 1]),
            ]


def test_none_2():

    @class_initialization_constraints
    class Case1(object):

        INIT_PARAMETERS = [
            Parameter('a', int, nullable=True, default=None),
        ]

    @class_initialization_constraints
    class Case2(object):

        INIT_PARAMETERS = [
            Parameter('a', type(None)),
        ]

    @class_initialization_constraints
    class Case3(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

    Case1()
    Case1(None)
    Case2(None)
    with pytest.raises(TypeError):
        Case3(None)


def test_corner_case():

    with pytest.raises(SyntaxError):
        function_constraints([])(1)
    with pytest.raises(SyntaxError):
        function_constraints(1)(1)

    with pytest.raises(SyntaxError):

        @class_initialization_constraints
        class Case2(object):

            INIT_PARAMETERS = [
                Parameter('duplicates', bool),
                Parameter('duplicates', bool),
            ]

    @class_initialization_constraints
    class Case3(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

    with pytest.raises(SyntaxError):
        Case3()
    with pytest.raises(SyntaxError):
        Case3(1, 2)
    with pytest.raises(SyntaxError):
        Case3(1, a=2)
    with pytest.raises(SyntaxError):
        Case3(b=2)

    with pytest.raises(TypeError):

        @class_initialization_constraints
        class Case4(object):

            INIT_PARAMETERS = [
                Parameter('a', 1),
            ]

    with pytest.raises(SyntaxError):

        @class_initialization_constraints
        class Case5(object):

            INIT_PARAMETERS = [
                Parameter('a', bool),
                Parameter('with_default', bool, default=None),
                Parameter('without_default', bool),
            ]


def test_types():

    @class_initialization_constraints
    class Case1(object):

        INIT_PARAMETERS = [
            Parameter('a', MutableSequence[int]),
        ]

    Case1(
        [1, 2, 3],
    )
    with pytest.raises(TypeError):
        Case1(
            [1, 2.0],
        )
    with pytest.raises(TypeError):
        Case1(
            (1, 2),
        )

    @class_initialization_constraints
    class Case2(object):

        INIT_PARAMETERS = [
            Parameter('a', ImmutableSequence[int]),
        ]

    Case2(
        (1, 2, 3),
    )
    with pytest.raises(TypeError):
        Case2(
            (1, 2.0),
        )
    with pytest.raises(TypeError):
        Case2(
            [1, 2],
        )

    @class_initialization_constraints
    class Case3(object):

        INIT_PARAMETERS = [
            Parameter(
                'a', Sequence[int],
                nullable=True, default=None,
            ),
            Parameter(
                'b', MutableSequence[Union[int, float]],
                nullable=True, default=None,
            ),
        ]

    Case3(
        a=[1, 2, 3],
    )
    Case3(
        a=(1, 2, 3),
    )
    Case3(
        b=[1, 2],
    )
    Case3(
        b=[1, 2.0],
    )
    with pytest.raises(TypeError):
        Case3(
            a=[1, 2.0],
        )
    with pytest.raises(TypeError):
        Case3(
            b=(1, 2.0),
        )



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from magic_constraints import *  # noqa
from magic_constraints.utils import (
    repr_return,
)


def test_exception():
    assert issubclass(MagicSyntaxError, MagicError)
    assert issubclass(MagicTypeError, MagicError)


def test_repr_and_str():

    assert (
        repr_return(
            "MagicSyntaxError('[Empty Message]',)"
        ) ==
        repr(MagicSyntaxError())
    )

    assert (
        repr_return(
            "MagicSyntaxError('msg',)"
        ) ==
        repr(MagicSyntaxError('msg'))
    )

    assert (
        repr_return(
            "MagicSyntaxError('msg',)"
        ) ==
        repr(MagicSyntaxError('msg', b=1, a=2))
    )

    assert (
        repr_return(
            "\n"
            "MagicSyntaxError: [Empty Message]\n"
            "---------------------------------\n"
            "[Empty kwargs]\n"
            "---------------------------------"
        ) ==
        str(MagicSyntaxError())
    )

    assert (
        repr_return(
            "\n"
            "MagicSyntaxError: msg\n"
            "---------------------\n"
            "[Empty kwargs]\n"
            "---------------------"
        ) ==
        str(MagicSyntaxError('msg'))
    )

    assert (
        repr_return(
            "\n"
            "MagicSyntaxError: msg\n"
            "---------------------\n"
            "a: 2\n"
            "b: 1\n"
            "---------------------"
        ) ==
        str(MagicSyntaxError('msg', b=1, a=2))
    )


def test_serialize():

    e = MagicTypeError()
    assert {'message': '[Empty Message]'} == e.serialize()

    e = MagicTypeError('msg')
    assert {'message': 'msg'} == e.serialize()

    e = MagicTypeError('msg', a=1, b=2)
    assert {
        'message': 'msg',
        'a': 1,
        'b': 2,
    } == e.serialize()



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function,
    # commented out since __all__ require str in Python 2.
    # unicode_literals,
)

from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from magic_constraints.exception import (
    MagicError,
    MagicSyntaxError,
    MagicTypeError,
)

from magic_constraints.constraint import Parameter, ReturnType

from magic_constraints.decorator import (
    function_constraints,
    method_constraints,
    class_initialization_constraints,
)

from magic_constraints.types import (
    Sequence,
    MutableSequence,
    ImmutableSequence,

    Set,
    MutableSet,
    ImmutableSet,

    Mapping,
    MutableMapping,
    ImmutableMapping,

    Iterator,
    Iterable,

    Any,
    Union,
)

__all__ = [
    'MagicError',
    'MagicSyntaxError',
    'MagicTypeError',

    'Parameter',
    'ReturnType',

    'function_constraints',
    'method_constraints',
    'class_initialization_constraints',

    'Sequence',
    'MutableSequence',
    'ImmutableSequence',

    'Set',
    'MutableSet',
    'ImmutableSet',

    'Mapping',
    'MutableMapping',
    'ImmutableMapping',

    'Iterator',
    'Iterable',

    'Any',
    'Union',
]



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import sys
import pytest
from magic_constraints import *  # noqa
from magic_constraints.utils import (
    repr_return,
)


pytestmark = pytest.mark.skipif(
    sys.version_info.major < 3,
    reason="future problem.",
)


def test_validator():
    p = Parameter('a', int, validator=lambda n: n == 42)
    assert p.check_argument(42)
    assert not p.check_argument(0)
    assert not p.check_argument('test')


def test_repr():
    assert (
        repr_return("Parameter(name='a', type_=Sequence[int])") ==
        repr(Parameter(name='a', type_=Sequence[int]))
    )
    assert (
        repr_return(
            "Parameter("
            "name='a', "
            "type_=Sequence[int], "
            "nullable=True"
            ")"
        ) ==
        repr(Parameter(name='a', type_=Sequence[int], nullable=True))
    )
    # test keyword sorting.
    assert (
        repr_return(
            "Parameter("
            "name='a', "
            "type_=Sequence[int], "
            "default=[1, 2, 3], "
            "nullable=True"
            ")"
        ) ==
        repr(Parameter(
            name='a', type_=Sequence[int], nullable=True, default=[1, 2, 3],
        ))
    )



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import sys
import pytest
from magic_constraints import *  # noqa
from magic_constraints.utils import (
    repr_return,
)


def test_sequence():

    assert isinstance([], Sequence)
    assert isinstance((1, 2, 3), Sequence[int])
    assert not isinstance([1, 2.0, 3], Sequence[int])

    assert isinstance([1, 2.0], Sequence[int, float])
    assert not isinstance([1, 2], Sequence[int, float])
    assert not isinstance([1], Sequence[int, float])
    assert not isinstance([1.0], Sequence[int, float])

    assert isinstance([], MutableSequence)
    assert isinstance([1, 2, 3], MutableSequence)
    assert not isinstance((1, 2, 3), MutableSequence)

    assert isinstance((), ImmutableSequence)
    assert isinstance((1, 2, 3), ImmutableSequence)
    assert not isinstance([1, 2, 3], ImmutableSequence)

    assert not issubclass(Sequence[int], Sequence)
    assert not issubclass(Sequence, Sequence[int])
    assert not isinstance(Sequence[int], Sequence)
    assert not isinstance(Sequence, Sequence[int])

    assert issubclass(MutableSequence, Sequence)
    assert issubclass(ImmutableSequence, Sequence)
    assert not issubclass(MutableSequence, ImmutableSequence)
    assert not issubclass(ImmutableSequence, MutableSequence)

    with pytest.raises(TypeError):
        Sequence[int, 1]

    with pytest.raises(TypeError):
        Sequence[[int, int]]


def test_set():

    assert isinstance({1, 2}, Set)
    assert isinstance({1, 2}, MutableSet)
    assert not isinstance({1, 2}, ImmutableSet)
    assert isinstance(frozenset((1, 2)), ImmutableSet)

    assert isinstance({1, 2}, Set[int])
    assert not isinstance({1, 2.0}, Set[int])

    with pytest.raises(TypeError):
        Set[int, ]
    with pytest.raises(TypeError):
        Set[int, float]

    assert issubclass(MutableSet, Set)
    assert issubclass(ImmutableSet, Set)
    assert not issubclass(MutableSet, ImmutableSet)
    assert not issubclass(ImmutableSet, MutableSet)


def test_mapping():
    assert isinstance({'a': 1}, Mapping)
    assert isinstance({'a': 1}, MutableMapping)

    assert isinstance({'a': 1}, Mapping[str, int])
    assert not isinstance({'a': 1.0}, Mapping[str, int])
    assert not isinstance({1: 1}, Mapping[str, int])

    assert isinstance({'a': 1}, MutableMapping[str, int])
    assert not isinstance({'a': 1.0}, MutableMapping[str, int])
    assert not isinstance({1: 1}, MutableMapping[str, int])

    assert not isinstance(1, Mapping)

    with pytest.raises(TypeError):
        Mapping[int]
    with pytest.raises(TypeError):
        Mapping[int, ]
    with pytest.raises(TypeError):
        Mapping[int, int, int]

    assert issubclass(MutableMapping, Mapping)
    assert issubclass(ImmutableMapping, Mapping)
    assert not issubclass(MutableMapping, ImmutableMapping)
    assert not issubclass(ImmutableMapping, MutableMapping)


@pytest.mark.skipif(
    sys.version_info < (3, 3),
    reason="types.MappingProxyType new in 3.3",
)
def test_immutable_mapping():
    from types import MappingProxyType as idict
    assert isinstance(
        idict({'a': 1}), ImmutableMapping,
    )

    assert isinstance(
        idict({'a': 1}), Mapping[str, int],
    )
    assert not isinstance(
        idict({'a': 1.0}), Mapping[str, int],
    )
    assert not isinstance(
        idict({1: 1}), Mapping[str, int],
    )


def test_iterator():
    assert isinstance(iter([1, 2]), Iterator)
    assert not isinstance([1, 2], Iterator)

    for _ in Iterator[int](iter([1, 2])):
        pass

    with pytest.raises(TypeError):
        Iterator([1, 2])
    with pytest.raises(TypeError):
        Iterator[int]([1, 2])

    with pytest.raises(TypeError):
        for _ in Iterator[int](iter([1, 2.0])):
            pass
    with pytest.raises(TypeError):
        Iterator[int, int]

    assert isinstance(
        (i for i in range(10)),
        Iterator,
    )


def test_iterable():
    assert isinstance([1, 2, 3], Iterable)
    assert not isinstance(1, Iterable)

    for _ in Iterable[int]([1, 2]):
        pass

    with pytest.raises(TypeError):
        Iterable(1)
    with pytest.raises(TypeError):
        Iterable[int](1)

    with pytest.raises(TypeError):
        for _ in Iterable[int]([1, 2.0]):
            pass
    with pytest.raises(TypeError):
        Iterable[int, int]

    assert isinstance(
        (i for i in range(10)),
        Iterable,
    )


def test_any():

    assert isinstance(1, Any)
    assert isinstance(1.0, Any)
    assert isinstance(object, Any)
    assert isinstance(type, Any)

    assert not issubclass(1, Any)
    assert not issubclass(1.0, Any)
    assert issubclass(object, Any)
    assert issubclass(type, Any)

    assert issubclass(Sequence, Any)
    assert issubclass(Union, Any)


def test_union():
    assert not isinstance(1, Union)
    assert not issubclass(int, Union)

    assert isinstance(1, Union[int, float])
    assert isinstance(1.0, Union[int, float])
    assert not isinstance('str', Union[int, float])

    assert not issubclass(str, Union)
    assert not issubclass(str, Union[int, float])

    with pytest.raises(TypeError):
        Union[1]


@pytest.mark.xfail(
    sys.version_info.major > 2,
    reason="future problem.",
)
def test_repr_python2():
    assert repr_return('Sequence') == repr(Sequence)
    assert repr_return('Sequence[newint]') == repr(Sequence[int])
    assert repr_return('Sequence[newint, float]') == repr(Sequence[int, float])


@pytest.mark.xfail(
    sys.version_info.major < 3,
    reason="future problem.",
)
def test_repr_python3():
    assert repr_return('Sequence') == repr(Sequence)
    assert repr_return('Sequence[int]') == repr(Sequence[int])
    assert repr_return('Sequence[int, float]') == repr(Sequence[int, float])


def test_corner_cases():
    assert not issubclass(1, Sequence)
    assert not issubclass(int, ImmutableSequence)
    assert not isinstance(int, ImmutableSequence)



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from magic_constraints.metadata import *  # noqa


def test_just_to_coverage():
    NAME
    VERSION
    AUTHORS
    EMAILS
    LICENSE
    URL
    DESCRIPTION



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import sys
import pytest
from magic_constraints import *  # noqa


pytestmark = pytest.mark.skipif(
    sys.version_info.major < 3,
    reason="",
)


def test_annotation():

    @function_constraints
    def func1(foo: int, bar: float) -> float:
        return foo + bar

    assert 3.0 == func1(1, 2.0)

    @function_constraints
    def func2(foo: int, bar: float = None) -> float:
        if bar is None:
            # should fail the return type checkin.
            return 42
        else:
            # good case.
            return foo + bar

    assert 3.0 == func2(1, 2.0)
    with pytest.raises(TypeError):
        func2(1)



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa


class MagicError(Exception):

    def __init__(self, message='[Empty Message]', **kwargs):
        super().__init__(repr_return(message))

        self._message = message
        self._kwargs = kwargs

    def __str__(self):
        title = '{exception_name}: {message}'.format(
            exception_name=type(self).__name__,
            message=self._message,
        )
        sep_line = '-' * len(title)

        kv_line_template = '{0}: {1}'
        kv_lines = '\n'.join(map(
            lambda p: kv_line_template.format(p[0], p[1]),
            sorted(self._kwargs.items(), key=lambda p: p[0]),
        ))

        return repr_return(
            # prefix \n for traceback display.
            '\n{title}\n{sep_line}\n{kv_lines}\n{sep_line}'.format(
                title=title,
                kv_lines=kv_lines or '[Empty kwargs]',
                sep_line=sep_line,
            ),
        )

    def serialize(self):
        ret = dict(self._kwargs)
        ret['message'] = self._message
        return ret


class MagicSyntaxError(MagicError, SyntaxError):
    pass


class MagicTypeError(MagicError, TypeError):
    pass


from magic_constraints.utils import repr_return  # noqa



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa
from future.utils import with_metaclass

from abc import ABCMeta
# collections.abc dosn't esist in Python 2.x.
import collections as abc

from magic_constraints.exception import MagicTypeError
from magic_constraints.utils import (
    type_object, nontype_object,
    repr_return, conditional_repr,
)


def unbound_getattr(target, name):
    return target.__dict__.get(name, None)


def CreateMetaMagicType(generator_cls):

    class MetaMagicType(ABCMeta):

        def __getitem__(cls, type_decl):

            type_decl_checker = unbound_getattr(
                generator_cls,
                'check_getitem_type_decl',
            )
            if type_decl_checker and not type_decl_checker(type_decl):
                raise MagicTypeError(
                    'invalid type.',
                    type_decl=type_decl,
                )

            ret_cls = generator_cls(cls.main_cls)
            ret_cls.partial_cls = type_decl
            return ret_cls

        # fix unbound error of Python 2.x.
        __instancecheck__ = unbound_getattr(generator_cls, '__instancecheck__')

        def __subclasscheck__(cls, subclass):
            if nontype_object(subclass):
                return False

            if unbound_getattr(generator_cls, '__subclasshook__'):
                subclasshook = unbound_getattr(
                    generator_cls, '__subclasshook__',
                )
                if not subclasshook(cls, subclass):
                    return False

            # corner case, subclass isn't MagicType.
            if not hasattr(subclass, 'partial_cls'):
                return issubclass(subclass, cls.main_cls)

            # subclass is MagicType.
            if cls.partial_cls or subclass.partial_cls:
                # if subclass has partial_cls, return False.
                return False
            else:
                # 1. subclass is normal type object.
                # 2. subclass is a MagicType.
                return issubclass(subclass.main_cls, cls.main_cls)

        def __repr__(cls):

            name = conditional_repr(cls.main_cls)
            if cls.partial_cls:
                partial = ', '.join(
                    map(
                        conditional_repr,
                        cls.partial_cls
                        if isinstance(cls.partial_cls, tuple)
                        else [cls.partial_cls],
                    ),
                )

                name = '{0}[{1}]'.format(
                    name, partial,
                )

            return repr_return(name)

    return MetaMagicType


def CreateMagicType(generator_cls, MetaMagicType, ABC):

    class MagicType(with_metaclass(MetaMagicType, object)):

        main_cls = ABC
        partial_cls = None

        for attr in [
            '__new__',
            '__init__',
            # iterator.
            '__iter__',
            '__next__',
        ]:
            if unbound_getattr(generator_cls, attr):
                locals()[attr] = unbound_getattr(generator_cls, attr)

    return MagicType


class MagicTypeGenerator(type):

    def __new__(cls, ABC):
        MetaMagicType = CreateMetaMagicType(cls)
        return CreateMagicType(cls, MetaMagicType, ABC)


def check_type_of_instance(cls, instance):
    return any(
        issubclass(t, cls)
        # handle old style class.
        for t in (instance.__class__, type(instance))
    )


def generate_immutable_abc(supercls, mutable_subclass):

    class ABCImmutableMeta(ABCMeta):

        def __subclasscheck__(cls, subclass):
            if not issubclass(subclass, supercls):
                return False
            return not issubclass(subclass, mutable_subclass)

    class ABCImmutable(with_metaclass(ABCImmutableMeta, object)):
        pass

    # dirty hack to assert issubclass(ABCImmutable, supercls).
    supercls._abc_cache.add(ABCImmutable)

    return ABCImmutable


class SequenceGenerator(MagicTypeGenerator):

    def check_getitem_type_decl(type_decl):
        if type_object(type_decl):
            return True

        if isinstance(type_decl, tuple):
            for t in type_decl:
                if nontype_object(t):
                    return False
            return True
        else:
            return False

    def __instancecheck__(cls, instance):
        if not check_type_of_instance(cls, instance):
            return False

        if not cls.partial_cls:
            return True

        if type_object(cls.partial_cls):
            for e in instance:
                if not isinstance(e, cls.partial_cls):
                    return False
        else:
            if len(cls.partial_cls) != len(instance):
                return False
            for i in range(len(instance)):
                if not isinstance(instance[i], cls.partial_cls[i]):
                    return False

        return True


class SetGenerator(MagicTypeGenerator):

    def check_getitem_type_decl(type_decl):
        return type_object(type_decl)

    def __instancecheck__(cls, instance):
        if not check_type_of_instance(cls, instance):
            return False

        if not cls.partial_cls:
            return True

        if type_object(cls.partial_cls):
            for e in instance:
                if not isinstance(e, cls.partial_cls):
                    return False

        return True


class MappingGenerator(MagicTypeGenerator):

    def check_getitem_type_decl(type_decl):
        if type_object(type_decl) or not isinstance(type_decl, tuple):
            return False
        if len(type_decl) != 2:
            return False
        return type_object(type_decl[0]) and type_object(type_decl[1])

    def __instancecheck__(cls, instance):
        if not check_type_of_instance(cls, instance):
            return False

        if cls.partial_cls:
            key_cls, val_cls = cls.partial_cls
            for key, val in instance.items():
                if not (isinstance(key, key_cls) and isinstance(val, val_cls)):
                    return False
        return True


class IteratorGenerator(MagicTypeGenerator):

    def check_getitem_type_decl(type_decl):
        return type_object(type_decl)

    def __init__(self, iterator):
        if self.partial_cls is None:
            raise MagicTypeError(
                'require T on Iterator[T].'
            )
        if not isinstance(iterator, abc.Iterator):
            raise MagicTypeError(
                'require Iterator.',
                iterator=iterator,
            )

        self.iterator = iterator

    def __iter__(self):
        return self

    def __next__(self):
        element = next(self.iterator)
        if isinstance(element, self.partial_cls):
            return element
        else:
            raise MagicTypeError(
                'type unmatched.',
                element=element,
                type_=self.partial_cls,
            )

    def __instancecheck__(cls, instance):
        return check_type_of_instance(cls, instance)


class IterableGenerator(MagicTypeGenerator):

    def check_getitem_type_decl(type_decl):
        return type_object(type_decl)

    def __init__(self, iterable):
        if self.partial_cls is None:
            raise MagicTypeError(
                'require T on Iterable[T].'
            )
        if not isinstance(iterable, abc.Iterable):
            raise MagicTypeError(
                'require Iterable.',
                iterable=iterable,
            )
        self.iterable = iterable

    def __iter__(self):
        return Iterator[self.partial_cls](
            iter(self.iterable),
        )

    def __instancecheck__(cls, instance):
        return check_type_of_instance(cls, instance)


class AnyMeta(ABCMeta):

    def __instancecheck__(cls, instance):
        return True

    def __subclasscheck__(cls, subclass):
        return type_object(subclass)


class Any(with_metaclass(AnyMeta, object)):
    pass


class UnionGenerator(MagicTypeGenerator):

    def check_getitem_type_decl(type_decl):
        if type_object(type_decl) or not isinstance(type_decl, tuple):
            return False

        for t in type_decl:
            if nontype_object(t):
                return False
        else:
            return True

    def __subclasshook__(cls, subclass):
        return False

    def __instancecheck__(cls, instance):
        if cls.partial_cls is None:
            return False

        for candidate_cls in cls.partial_cls:
            if isinstance(instance, candidate_cls):
                return True
        return False


ABCImmutableSequence = generate_immutable_abc(
    abc.Sequence, abc.MutableSequence,
)
ABCImmutableSet = generate_immutable_abc(
    abc.Set, abc.MutableSet,
)
ABCImmutableMapping = generate_immutable_abc(
    abc.Mapping, abc.MutableMapping,
)

Sequence = SequenceGenerator(abc.Sequence)
MutableSequence = SequenceGenerator(abc.MutableSequence)
ImmutableSequence = SequenceGenerator(ABCImmutableSequence)

Set = SetGenerator(abc.Set)
MutableSet = SetGenerator(abc.MutableSet)
ImmutableSet = SetGenerator(ABCImmutableSet)

Mapping = MappingGenerator(abc.Mapping)
MutableMapping = MappingGenerator(abc.MutableMapping)
ImmutableMapping = MappingGenerator(ABCImmutableMapping)

Iterator = IteratorGenerator(abc.Iterator)
Iterable = IterableGenerator(abc.Iterable)

Union = UnionGenerator(Any)



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import pytest
from magic_constraints import *  # noqa


def test_function_constraints_pass_by_compound():

    @function_constraints(
        Parameter('a', int),
        Parameter('b', float),
    )
    def example(args):
        assert args.a == 1
        assert args.b == 1.0

    example(1, 1.0)
    example(b=1.0, a=1)
    with pytest.raises(AssertionError):
        example(2, 1.0)

    with pytest.raises(SyntaxError):
        example(1)
    with pytest.raises(SyntaxError):
        example(1.0)
    with pytest.raises(SyntaxError):
        example(1, a=1)


def test_function_constraints_pass_by_positional():

    @function_constraints(
        int, float, int, int,
    )
    def example(a, b, c=42, d=None):
        assert a == 1
        assert b == 1.0
        assert c == 42
        assert d is None

    example(1, 1.0)
    example(b=1.0, a=1)
    with pytest.raises(AssertionError):
        example(2, 1.0)

    example(1, 1.0, 42)
    with pytest.raises(AssertionError):
        example(2, 1.0, 43)

    example(1, 1.0, 42, None)
    with pytest.raises(AssertionError):
        example(1, 1.0, 42, d=42)

    with pytest.raises(SyntaxError):
        example(1)
    with pytest.raises(SyntaxError):
        example(1.0)
    with pytest.raises(SyntaxError):
        example(1, a=1)

    with pytest.raises(SyntaxError):
        function_constraints()


def test_method_constraints():

    class Case1(object):

        @classmethod
        @method_constraints(
            Parameter('a', int),
        )
        def test_cls(cls, args):
            assert args.a == 1

        @method_constraints(
            int,
        )
        def test_self(self, a):
            assert a == 1

    Case1.test_cls(1)
    Case1.test_cls(a=1)
    with pytest.raises(AssertionError):
        Case1.test_cls(2)

    case1 = Case1()
    case1.test_self(1)
    with pytest.raises(AssertionError):
        case1.test_self(2)

    with pytest.raises(SyntaxError):
        method_constraints()


def test_class_initialization_constraints():

    @class_initialization_constraints
    class Case1(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

        def __init__(self):
            assert self.a == 1

    Case1(1)
    Case1(a=1)
    with pytest.raises(AssertionError):
        Case1(2)

    @class_initialization_constraints
    class Case2(object):

        INIT_PARAMETERS = [
            Parameter('a', int),
        ]

    Case2(1)
    Case2(a=1)

    with pytest.raises(TypeError):
        class_initialization_constraints(1)


def test_corner_cases1():

    with pytest.raises(SyntaxError):
        @function_constraints(
            list(),
        )
        def example(args):
            pass

    with pytest.raises(TypeError):
        @class_initialization_constraints
        class Case2(object):

            INIT_PARAMETERS = 42

    with pytest.raises(SyntaxError):
        @function_constraints(
            int, float,
        )
        def func1(a):
            pass

    with pytest.raises(SyntaxError):
        @function_constraints(
            int, float,
        )
        def func2(*a, **b):
            pass


def test_return_type():

    @function_constraints(
        int,
        return_type=int,
    )
    def func1(a):
        return a

    @function_constraints(
        int,
        return_type=float,
    )
    def func2(a):
        return a

    func1(1)
    with pytest.raises(TypeError):
        func2(1)

    @function_constraints(
        Parameter('a', int, nullable=True),
        ReturnType(float, nullable=True),
    )
    def func3(args):
        return args.a

    func3(None)
    with pytest.raises(TypeError):
        func3(1)

    class Example(object):

        @method_constraints(
            int,
            return_type=float,
        )
        def func1(self, a):
            return a

        @method_constraints(
            Parameter('a', int, nullable=True),
            ReturnType(float, nullable=True),
        )
        def func2(self, args):
            return args.a

    e = Example()
    with pytest.raises(TypeError):
        e.func1(1)
    e.func2(None)
    with pytest.raises(TypeError):
        e.func2(1)

    @function_constraints(
        ReturnType(float, nullable=True),
    )
    def func4(args):
        return None

    func4()

    @function_constraints(
        return_type=int
    )
    def func5():
        return 42

    func5()



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa
import types

from magic_constraints.exception import (
    MagicSyntaxError,
    MagicTypeError,
)

from magic_constraints.constraint import (
    Constraint,

    build_constraints_package,
    build_constraints_with_given_type_args,
    build_constraints_with_annotation,

    raise_on_non_constraints,
    raise_on_non_parameters,
)

from magic_constraints.argument import (
    transform_to_slots, check_and_bind_arguments,
)

from magic_constraints.utils import (
    CompoundArgument,
    AttributesBinder,
    type_object,
    raise_on_non_function,
    raise_on_nontype_object,
)


def decorator_dispather(args, options,
                        by_positional, by_compound, by_inspection):
    if not args and 'return_type' not in options:
        raise MagicSyntaxError(
            'empty args with no return_type option.',
        )

    if 'return_type' in options or type_object(args[0]):
        return by_positional(args, options)

    elif isinstance(args[0], Constraint):
        return by_compound(args, options)

    elif len(args) == 1 and isinstance(args[0], types.FunctionType):
        return by_inspection(args[0])

    else:
        raise MagicSyntaxError(
            'can not dispatch.',
        )


def check_ret(ret, return_type):
    if not return_type.check_argument(ret):
        raise MagicTypeError(
            'return value unmatched.',
            return_type=return_type,
            ret=ret,
        )


# @function_constraints(
#     int, float,
#     return_type=xxx,
# )
# def function(foo, bar=1.0):
#     return foo + bar
#
#
def _function_constraints_pass_by_positional_args(type_args, options):

    def decorator(function):
        raise_on_non_function(function)

        input_type_args = [function, False, type_args]
        if 'return_type' in options:
            input_type_args.append(options['return_type'])

        constraints_package = build_constraints_package(
            build_constraints_with_given_type_args(*input_type_args),
        )

        def wrapper(*args, **kwargs):

            slots = transform_to_slots(constraints_package, *args, **kwargs)
            check_and_bind_arguments(
                constraints_package.parameters, slots, lambda name, arg: None,
            )

            ret = function(*slots)
            check_ret(ret, constraints_package.return_type)
            return ret

        return wrapper
    return decorator


# @function_constraints(
#     Parameter('foo', int),
#     Parameter('bar', float, default=1.0),
#     ...,
#     ReturnType(float),
# )
# def function(args):
#     return args.foo + args.bar
#
#
def _function_constraints_pass_by_compound_args(constraints, options):
    raise_on_non_constraints(constraints)

    constraints_package = build_constraints_package(constraints)

    def decorator(function):
        raise_on_non_function(function)

        def wrapper(*args, **kwargs):

            slots = transform_to_slots(constraints_package, *args, **kwargs)

            compound_args = CompoundArgument()
            bind_callback = AttributesBinder(compound_args)

            check_and_bind_arguments(
                constraints_package.parameters, slots, bind_callback,
            )

            ret = function(compound_args)
            check_ret(ret, constraints_package.return_type)
            return ret

        return wrapper
    return decorator


# @function_constraints
# def function(foo: int, bar: float) -> float:
#     return foo + bar
def _function_constraints_by_inspection(function):
    raise_on_non_function(function)

    constraints_package = build_constraints_package(
        build_constraints_with_annotation(function, False),
    )

    def wrapper(*args, **kwargs):
        slots = transform_to_slots(constraints_package, *args, **kwargs)
        check_and_bind_arguments(
            constraints_package.parameters, slots, lambda name, arg: None,
        )

        ret = function(*slots)
        check_ret(ret, constraints_package.return_type)
        return ret

    return wrapper


def function_constraints(*args, **options):
    return decorator_dispather(
        args, options,
        _function_constraints_pass_by_positional_args,
        _function_constraints_pass_by_compound_args,
        _function_constraints_by_inspection,
    )


# @method_constraints(
#     int, float,
#     return_type=xxx,
# )
# def method(self_or_cls, foo, bar):
#     return foo + bar
#
#
def _method_constraints_pass_by_positional_args(type_args, options):

    def decorator(function):
        raise_on_non_function(function)

        input_type_args = [function, True, type_args]
        if 'return_type' in options:
            input_type_args.append(options['return_type'])

        constraints_package = build_constraints_package(
            build_constraints_with_given_type_args(*input_type_args),
        )

        def wrapper(self_or_cls, *args, **kwargs):

            slots = transform_to_slots(constraints_package, *args, **kwargs)
            check_and_bind_arguments(
                constraints_package.parameters, slots, lambda name, arg: None,
            )

            ret = function(self_or_cls, *slots)
            check_ret(ret, constraints_package.return_type)
            return ret

        return wrapper
    return decorator


# @method_constraints(
#     Parameter('foo', int),
#     Parameter('bar', float, default=1.0),
#     ...,
#     ReturnType(float),
# )
# def method(self_or_cls, args):
#     return args.foo + args.bar
#
#
def _method_constraints_pass_by_compound_args(constraints, options):
    raise_on_non_constraints(constraints)

    constraints_package = build_constraints_package(constraints)

    def decorator(function):
        raise_on_non_function(function)

        def wrapper(self_or_cls, *args, **kwargs):

            slots = transform_to_slots(constraints_package, *args, **kwargs)

            compound_args = CompoundArgument()
            bind_callback = AttributesBinder(compound_args)

            check_and_bind_arguments(
                constraints_package.parameters, slots, bind_callback,
            )

            ret = function(self_or_cls, compound_args)
            check_ret(ret, constraints_package.return_type)
            return ret

        return wrapper
    return decorator


# @method_constraints
# def method(self_or_cls, foo: int, bar: float) -> float:
#     return foo + bar
def _method_constraints_by_inspection(function):
    raise_on_non_function(function)

    constraints_package = build_constraints_package(
        build_constraints_with_annotation(function, True),
    )

    def wrapper(self_or_cls, *args, **kwargs):
        slots = transform_to_slots(constraints_package, *args, **kwargs)
        check_and_bind_arguments(
            constraints_package.parameters, slots, lambda name, arg: None,
        )

        ret = function(self_or_cls, *slots)
        check_ret(ret, constraints_package.return_type)
        return ret

    return wrapper


def method_constraints(*args, **options):
    return decorator_dispather(
        args, options,
        _method_constraints_pass_by_positional_args,
        _method_constraints_pass_by_compound_args,
        _method_constraints_by_inspection,
    )


# @class_initialization_constraints
# class FooBar(object):
#
#     INIT_PARAMETERS = [
#         Parameter('foo', int),
#         Parameter('bar', float, default=1.0),
#         ...,
#         ReturnType(float),
#     ]
#
#
def class_initialization_constraints(user_defined_class):
    raise_on_nontype_object(user_defined_class)

    parameters = getattr(user_defined_class, 'INIT_PARAMETERS', None)
    raise_on_non_parameters(parameters)
    constraints_package = build_constraints_package(parameters)

    predefined_init = getattr(
        user_defined_class,
        '__init__',
    )

    def init(self, *args, **kwargs):

        slots = transform_to_slots(constraints_package, *args, **kwargs)
        bind_callback = AttributesBinder(self)
        check_and_bind_arguments(
            parameters, slots, bind_callback,
        )

        predefined_init(self)

    setattr(user_defined_class, '__init__', init)

    return user_defined_class
